/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * A convenient place to store all global configuration.
 *
 * $Id: GlobalConfig.hh,v 1.26 2007-10-19 17:49:17 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _GLOBAL_CONFIG_HH_
#define _GLOBAL_CONFIG_HH_

#include <string>
#include <vector>
#include <sys/types.h>
#include "Enums.hh"
#include "Exceptions.hh"
#include "Mutex.hh"

namespace HTTPTunnel {

class Authenticator;
class Server;
class SideConfig;

class GlobalConfig {
public:

	/*********************/
	/* Types & Constants */
	/*********************/

	// Maximum size of local hostname
	static const unsigned int		OUR_HOST_NAME_MAX		= 64;


   // Construction/destruction
   GlobalConfig();
   ~GlobalConfig();


	/*************/
	/* Utilities */
	/*************/

	// Serialize to a config file
	void dumpToFile(std::string name) throw(ConfigError);

	// Perform any required late-configuration
	void doLateConfiguration(void) throw(ConfigError);


	/**********************/
	/* Accessors/mutators */
	/**********************/

	// Client/server settings
	void configureServer(bool ssl, std::string cert, std::string certKey,
		std::string lhost, unsigned int lport, std::string firewallIPs,
		std::string firewallPorts, bool auth) throw(AuthError);

	void reconfigureServer(unsigned int no, bool ssl, std::string cert,
		std::string certKey, std::string lhost, unsigned int lport,
		std::string firewallIPs, std::string firewallPorts, bool auth)
		throw(TunnelError);

	unsigned int configureDefaultServer(void) throw(AuthError);

	void configureClient(Enums::Protocol prot, bool ssl, std::string lhost,
		unsigned int lport, std::string shost, unsigned int sport, std::string suser,
		std::string spasswd, std::string dhost, unsigned int dport);
	void configureClient(Enums::Protocol prot, bool ssl, std::string lhost,
		unsigned int lport, std::string shost, unsigned int sport, std::string suser,
		std::string spasswd, std::string dhost, unsigned int dport, std::string phost,
		unsigned int pport, std::string puser, std::string ppasswd);
	void configureClient(Enums::Protocol prot, bool ssl, std::string lhost,
		unsigned int lport, std::string shost, unsigned int sport, std::string suser,
		std::string spasswd);
	void configureClient(Enums::Protocol prot, bool ssl, std::string lhost,
		unsigned int lport, std::string shost, unsigned int sport, std::string suser,
		std::string spasswd, std::string phost, unsigned int pport, std::string puser,
		std::string ppasswd);

	void reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
		std::string lhost, unsigned int lport, std::string shost, unsigned int
		sport, std::string suser, std::string spasswd, std::string dhost,
		unsigned int dport) throw(TunnelError);
	void reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
		std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
		std::string suser, std::string spasswd, std::string dhost, unsigned int dport,
		std::string phost, unsigned int pport, std::string puser, std::string ppasswd)
		throw(TunnelError);
	void reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
		std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
		std::string suser, std::string spasswd) throw(TunnelError);
	void reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
		std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
		std::string suser, std::string spasswd, std::string phost, unsigned int pport,
		std::string puser, std::string ppasswd) throw(TunnelError);

	unsigned int configureDefaultClient(void);


	// Get pseudo unique local hostname
	std::string getHostName(void) const;

	// Get/set daemon flag
	void setDaemon(bool flag);
	bool getDaemon(void) const;

	// Get/set debug level
	void setDebug(unsigned int debug);
	unsigned int getDebug(void) const;

	// Get/set tunnel testing flag
	void setTesting(bool flag);
	bool getTesting(void) const;

	// Process values
	unsigned int getPid(void) const;
	void setPIDFile(std::string);
	std::string getPIDFile(void) const;

	// Default realm
	std::string getRealm(void) const;
	void setRealm(std::string realm);

	// Data log file stem
	std::string getDataLogFileStem(void) const;
	void setDataLogFileStem(std::string stem);

	// Create/destroy local servers
	void createLocalServers(void) throw(TunnelError);
	void createNamedLocalServer(unsigned int no) throw(TunnelError);
	void destroyLocalServers(void);
	void destroyNamedLocalServer(unsigned int no);

	// Have we got at least one client or server configured?
	bool confCliSvr(void);


private:

	/********/
	/* Data */
	/********/

	// Mutex
	Threads::Mutex					m_mtx;

	// Should we daemonise?
	bool								m_daemon;

	// Should we do tunnel testing?
	bool								m_testing;

	// Pseudo unique local host identifer
	std::string						m_hostName;

	// Debug level
	unsigned int					m_debugLevel;

	// Our process ID
	pid_t								m_pid;
	std::string						m_pidFile;

	// Default realm
	std::string						m_realm;

	// Local servers and their config
	std::vector<Server*>			m_servers;
	std::vector<SideConfig*>	m_serverConfigs;

	// Data log file stem
	std::string						m_dataLogStem;

};


}	// namespace HTTPTunnel

#endif

