/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an SSL socket.
 *
 * $Id: SSLSock.cc,v 1.7 2007-09-27 17:49:50 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifdef USE_SSL

#include <ssl.h>
#include <err.h>
#include <errno.h>
#include "SSLSock.hh"
#include "Logger.hh"
#include "GlobalConfig.hh"
#include "Debug.hh"
#include "Utils.hh"
#include "Tunneler.hh"


// Global references
extern HTTPTunnel::Logger* g_pLogger;
extern HTTPTunnel::GlobalConfig* g_pConfig;


namespace HTTPTunnel {


// Initialise statics
bool SSLSock::c_initSSL = false;
bool SSLSock::c_shutdownSSL = false;


// Construction/destruction
SSLSock::SSLSock() throw(SockError)
: Sock(), m_ctx(0), m_abio(0), m_sbio(0), m_ssl(0), m_nonBlocking(false), m_cert(""),
  m_certKey(""), m_connectRetry(0)
{
}

SSLSock::SSLSock(std::string cert, std::string key) throw(SockError)
: Sock(), m_ctx(0), m_abio(0), m_sbio(0), m_ssl(0), m_nonBlocking(false), m_cert(cert),
  m_certKey(key), m_connectRetry(0)
{
}

SSLSock::SSLSock(int fd, std::string cliAddr, BIO* sbio, SSL* ssl) throw(SockError)
: Sock(fd, cliAddr), m_ctx(0), m_abio(0), m_sbio(sbio), m_ssl(ssl), m_nonBlocking(false),
  m_cert(""), m_certKey(""), m_connectRetry(0)
{
}

SSLSock::~SSLSock()
{
	DPRINTV("SSLSock::~SSLSock()");

	Threads::MutexVar mv(m_mtx);

	try {
		// Close SSL socket
		close();
	}
	catch(SockError& e) {
		g_pLogger->info(e.what());
	}
}


/********************/
/* Public Interface */
/********************/


// Wait for a new incoming SSL connection
void SSLSock::newConnection(Tunneler* tunneler) throw(SockError)
{
	Threads::MutexVar mv(m_mtx);

	// Socket ready to accept
	Sock* connectedSock = accept();

	try {
		// Tell tunneler
		tunneler->newConnectionArrived(connectedSock);
	}
	catch(TunnelError& e) {
		g_pLogger->info(e.what());
		delete connectedSock;
	}
}


// Set socket listening
void SSLSock::listen(void) throw(SockError)
{
	// SSL listen() functions performed by bind(), so nothing more to do
}


// Wait for a new incoming connection
Sock* SSLSock::accept(void) throw(SockError)
{
	Threads::MutexVar mv(m_mtx);

	// Try to accept connection via the BIO
	long res = BIO_do_accept(m_abio);
	if(res <= 0) {
		throw SockError("Failed to accept SSL connection", 0);
	}

	// Get SSL BIO for connection
	BIO* sbio = BIO_pop(m_abio);
	if(sbio == NULL) {
		failedAccept(sbio);
		throw SockError("Failed to accept SSL connection", 0);
	}

	// Get socket descriptor from SSL BIO
	int fd = -1;
	BIO_get_fd(sbio, &fd);

	// Get underlying SSL object
	SSL* ssl;
	BIO_get_ssl(sbio, &ssl);
	if(ssl == NULL) {
		failedAccept(sbio);
		throw SockError("Failed to accept SSL connection", 0);
	}

	// Set SSL options
	SSL_set_verify(ssl, SSL_VERIFY_NONE, NULL);

	// Do the SSL accept
	res = SSL_accept(ssl);
	if(res <= 0) {
		failedAccept(sbio);
		throw SockError("Failed to accept SSL connection", 0);
	}

	// Find client address
	std::string clientAddr = Utils::getPeerName(fd);
	DPRINTI("SSLSock::accept() - Got new connection from " + clientAddr);

	// Create and return new SSLSock to handle this connected socket
	return new SSLSock(fd, clientAddr, sbio, ssl);
}


// Establish connection out
void SSLSock::connect(std::string addr, unsigned int port)
	throw(SockError, CommsError)
{
	Threads::MutexVar mv(m_mtx);

	// Create SSL context
	m_ctx = SSL_CTX_new(SSLv23_client_method());
	if(m_ctx == NULL) {
		throw ConnectError("Failed to create connecting SSL context");
	}

	// Create BIO
	m_sbio = BIO_new_ssl_connect(m_ctx);
	if(m_sbio == NULL) {
		throw ConnectError("Failed to create connecting SSL bio");
	}

	// Setup SSL
	BIO_get_ssl(m_sbio, &m_ssl);
	SSL_set_mode(m_ssl, SSL_MODE_AUTO_RETRY);
	SSL_set_verify(m_ssl, SSL_VERIFY_NONE, NULL);

	// Non-blocking connect?
	if(m_nonBlocking) {
		// Request non-blocking IO on BIO object
		if(BIO_set_nbio(m_sbio, 1) <= 0) {
			//DPRINTE("BIO_set_nbio() in connect() returned error");
			throw ConnectError("Unable to set SSL socket nonblocking flag");
		}
	}

	// Create address of the form addr:port
	std::stringstream addrStr;
	addrStr << addr << ":" << port;
	BIO_set_conn_hostname(m_sbio, addrStr.str().c_str());

	// Try the connect
	long res = BIO_do_connect(m_sbio);
	if(res <= 0) {
		// Do we need to check BIO_should_retry() for non-blocking IO here?
		if(BIO_should_retry(m_sbio)) {
			// Non-blocking connect not yet completed
			DPRINTI("BIO_should_retry() said yes");
			throw SockError("Failed to connect to SSL server", EWOULDBLOCK);
		}
		else {
			DPRINTI("BIO_should_retry() said no");
			throw ConnectError("Failed to connect to SSL server");
		}
	}
	else {
		// Get socket descriptor from BIO
		BIO_get_fd(m_sbio, &m_fd);
	}
}


// Read from socket
int SSLSock::read(unsigned char* buff, size_t buffSize)
	throw(SockError, CommsError)
{
	Threads::MutexVar mv(m_mtx);

	// Read data from SSL
again:
	int noread = SSL_read(m_ssl, buff, buffSize);
	if(noread < 0) {
		int sslErr = SSL_get_error(m_ssl, noread);
		if(sslErr == SSL_ERROR_WANT_READ || sslErr == SSL_ERROR_WANT_WRITE)
			goto again;
		else
			throw SockError("Read error from SSL socket", 0);
	}
	else if(noread == 0) {
		// EOF
		throw ConnectionClosed("Other end closed SSL socket");
	}

	return noread;
}


// Write to socket
int SSLSock::write(const unsigned char* buff, size_t buffSize)
	throw(SockError, CommsError)
{
	Threads::MutexVar mv(m_mtx);

	// Write data to SSL
again:
	int nowritten = SSL_write(m_ssl, buff, buffSize);
	if(nowritten <= 0) {
		int sslErr = SSL_get_error(m_ssl, nowritten);
		if(sslErr == SSL_ERROR_WANT_READ || sslErr == SSL_ERROR_WANT_WRITE)
			goto again;
		else
			throw SockError("Write error to SSL socket", 0);
	}

	return nowritten;
}


// Get the result of a non-blocking connect attempt
void SSLSock::getConnectResult(bool& resultAvailable, bool& connected)
	throw(SockError)
{
	Threads::MutexVar mv(m_mtx);

	// Try the connect again
	long res = BIO_do_connect(m_sbio);
	if(res <= 0) {
		#ifdef DEBUG_TUN
		std::stringstream nfo;
		nfo << "BIO_do_connect() in getConnectResult() returned error " << res;
		DPRINTI(nfo.str());
		#endif

		// Shuld we retry again?
		if(BIO_should_retry(m_sbio)) {
			// Non-blocking connect not yet completed
			DPRINTI("BIO_should_retry() said yes");
			m_connectRetry++;

			// Have we tried enough?
			if(m_connectRetry > MAX_CONNECT_RETRY) {
				DPRINTI("Retry limit exceeded");
				resultAvailable = true;
				connected = false;
			}
			else {
				resultAvailable = false;
			}
		}
		else {
			// Connect must have failed
			DPRINTI("BIO_should_retry() said no");
			resultAvailable = true;
			connected = false;
		}
	}
	else {
		// Connect succeeded
		resultAvailable = true;
		connected = true;

		// Get socket descriptor from BIO
		BIO_get_fd(m_sbio, &m_fd);
	}
}


// Bind socket to an address
void SSLSock::bind(std::string addr, unsigned int port) throw(SockError)
{
	Threads::MutexVar mv(m_mtx);

	// Create address of the form addr:port
	std::stringstream addrStr;
	if(addr.size() > 0)
		addrStr << addr << ":";
	addrStr << port;

	// Setup SSL context
	m_ctx = SSL_CTX_new(SSLv23_server_method());
	if(m_ctx == NULL) {
		throw SockError("Failed to create SSL context", 0);
	}

	// Load certificates
	DPRINTV("Loading SSL certificate: " + m_cert);
	long res = SSL_CTX_use_certificate_chain_file(m_ctx, m_cert.c_str());
	if(res <= 0) {
		std::stringstream err;
		err << "Loading certs in bind() returned error " << res;
		g_pLogger->error(err.str());
		throw SockError("Failed to load certificates for SSL context", 0);
	}

	// Load private key
	DPRINTV("Loading SSL certificate key: " + m_certKey);
	res = SSL_CTX_use_PrivateKey_file(m_ctx, m_certKey.c_str(), SSL_FILETYPE_PEM);
	if(res <= 0) {
		std::stringstream err;
		err << "Loading private key in bind() returned error " << res;
		g_pLogger->error(err.str());
		throw SockError("Failed to load private key for SSL context", 0);
	}

	// Create accept BIO
	m_abio = BIO_new_accept((char*)(addrStr.str().c_str()));
	if(m_abio == NULL) {
		throw SockError("Failed to create SSL accept BIO", 0);
	}

	// Reuse address
	res = BIO_set_bind_mode(m_abio, BIO_BIND_REUSEADDR);
	if(res <= 0) {
		throw SockError("Failed to set REUSEADDR for SSL socket", 0);
	}

	// Create SSL BIO
	BIO* sbio = BIO_new_ssl(m_ctx, 0);
	if(sbio == NULL) {
		throw SockError("Failed to create SSL SSL BIO", 0);
	}

	// Insert SSL BIO into chain
	BIO_set_accept_bios(m_abio, sbio);

	// Bind
	res = BIO_do_accept(m_abio);
	if(res <= 0) {
		throw SockError("Failed to bind accepting SSL socket", 0);
	}

	// Get underlying socket descriptor from accept BIO object
	BIO_get_fd(m_abio, &m_fd);
}


// Turn on non-blocking mode for this socket
void SSLSock::nonblocking(void) throw(SockError)
{
	Threads::MutexVar mv(m_mtx);

	// Set non-blocking flag
	m_nonBlocking = true;
}


// Initialise SSL library
void SSLSock::initSSL(void) throw(CommsError)
{
	if(!c_initSSL) {
		c_initSSL = true;

		// Initialise SSL
		SSL_load_error_strings();
		SSL_library_init();

		// Seed RNG
	}
}

// Close SSL library
void SSLSock::shutdownSSL(void) throw(CommsError)
{
	if(!c_shutdownSSL) {
		c_shutdownSSL = true;
	}
}


/***********************/
/* Protected Interface */
/***********************/


// Close SSL session
void SSLSock::close(void) throw(SockError)
{
	DPRINTV("SSLSock::close()");

	Threads::MutexVar mv(m_mtx);

	// Free SSL BIO object and associated SSL object
	if(m_sbio) {
		BIO_free_all(m_sbio);
		m_sbio = 0;
		m_ssl = 0;
	}

	// Free accept BIO object and associated SSL object
	if(m_abio) {
		BIO_free_all(m_abio);
		m_abio = 0;
		m_ssl = 0;
	}

	// Free context
	if(m_ctx) {
		SSL_CTX_free(m_ctx);
		m_ctx = 0;
	}
}


/*********************/
/* Private Interface */
/*********************/


// Tidy up after a failed accept
void SSLSock::failedAccept(BIO* sbio)
{
	// Free SSL BIO object and associated SSL object
	if(sbio) {
		BIO_free_all(sbio);
		sbio = 0;
	}

	// Clear error queue
	ERR_clear_error();
}


}	// namespace

#endif	// Using SSL

