/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * A TCP/IP socket server to listen for incoming connections.
 *
 * $Id: Server.cc,v 1.29 2007-08-18 13:36:26 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <errno.h>
#include <poll.h>
#include "Server.hh"
#include "Debug.hh"
#include "Enums.hh"
#include "SideConfig.hh"
#include "ServerSideConfig.hh"
#include "Callback.hh"
#include "Exceptions.hh"
#include "Tunneler.hh"
#include "Logger.hh"
#include "MutexVar.hh"
#include "TCPSock.hh"
#include "UDPSock.hh"
#include "SSLSock.hh"


// Global references
extern HTTPTunnel::Logger* g_pLogger;


namespace HTTPTunnel {


// Construction/destruction
Server::Server(SideConfig* serverConfig) throw(TunnelError)
: m_thread(0), m_serverConfig(serverConfig), m_sock(0), m_tunneler(0),
  m_sleepCond(), m_shutdownMtx(), m_shutdownCond(), m_shutDown(false),
  m_threadMgr(), m_mtx()
{
}

Server::~Server()
{
	// Take mutexes
	Threads::MutexVar mv(m_mtx);
	Threads::MutexVar smv(m_shutdownMtx);

	// Signal we are no longer running
	m_shutDown = true;

	// Wakeup server thread & wait for it to finish
	if(m_thread) {
		m_sleepCond.signal();
		mv.unlock();
		m_shutdownCond.wait(smv);
		mv.lock();
		m_thread = 0;
	}

	// Finished with listening socket
	if(m_sock) {
		delete m_sock;
		m_sock = 0;
	}

	// Delete tunneler
	delete m_tunneler;
	m_tunneler = 0;
}


// Get config settings
const SideConfig* Server::getConfig(void) const
{
	return m_serverConfig;
}

SideConfig* Server::getConfigMutable(void)
{
	return m_serverConfig;
}


/***********************/
/* Protected interface */
/***********************/


// Wait for arriving connections (never returns)
void Server::startServer(void)
{
	DPRINTI("Server::startServer() - Waiting for connection");

	Threads::MutexVar mv(m_mtx);

	while(!m_shutDown) {
		try {
			// Poll socket until we think it's worth trying to accept()
			bool pollError, pollHup;
			if(m_sock->pollRead(0, pollError, pollHup)) {
				// Socket connection pending
				m_sock->newConnection(m_tunneler);
			}
			else {
				// Poll timeout, wait a while before retrying
				Time pollTime(0, POLL_TIMEOUT);
				m_sleepCond.timedwait(mv, pollTime.getTimeVal());
			}
		}
		catch(SockError& e) {
			if(e.m_errno == EINTR) {
				// Interrupted system call
				continue;
			}
			else {
				// General error, just carry on
				continue;
			}
		}
	}

	// Signal we have finished
	mv.unlock();
	Threads::MutexVar smv(m_shutdownMtx);
	m_shutdownCond.signal();
	smv.unlock();

	DPRINTI("Server closing down");
}


// Configure our server socket
void Server::configureSocket(bool ssl) throw(TunnelError)
{
	// Create listening socket
	if(ssl) {
		// Use SSL
		if(m_serverConfig->isServer()) {
			// Need both SSL certificate & private key at server
			ServerSideConfig* scfg = dynamic_cast<ServerSideConfig*>(m_serverConfig);
			m_sock = new SSLSock(scfg->getCertificate(), scfg->getCertificateKey());
		}
		else {
			// Don't require anything at client
			m_sock = new SSLSock();
		}
	}
	else {
		// Unencypted comms
		if(m_serverConfig->getProtocol() == Enums::TCP) {
			m_sock = new TCPSock();
		}
		else {
			m_sock = new UDPSock();
		}
	}

	// Setup listening socket
	m_sock->bind(m_serverConfig->getListenIP(), m_serverConfig->getListenPort());
	m_sock->listen();

	// Launch server
	Threads::ZeroArgClassCallback<Server> cb(&Server::startServer, this);
	m_thread = m_threadMgr.newThread(cb);
}


}	// namespace

