/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Wrap a socket connection in HTTP and tunnel it to some destination.
 *
 * $Id: main.cc,v 1.43 2007-10-19 17:49:17 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <sys/stat.h>
#include "GlobalConfig.hh"
#include "Exceptions.hh"
#include "ThreadExceptions.hh"
#include "Debug.hh"
#include "Logger.hh"
#include "Server.hh"
#include "OptionsParser.hh"
#include "SSLSock.hh"


using namespace HTTPTunnel;


/********/
/* Data */
/********/


// The global configuration
GlobalConfig*			g_pConfig = 0;

// A global logger
Logger*					g_pLogger = 0;

// Global flag to indicate we should keep running
bool						g_bRunning = true;


#ifdef DEBUG_LOG_DATA
// Global log files + Mutex
std::ofstream*			g_pRawReadLogFile = 0;
std::ofstream*			g_pRawWriteLogFile = 0;
std::ofstream*			g_pPayloadReadLogFile = 0;
std::ofstream*			g_pPayloadWriteLogFile = 0;
unsigned int			g_RawReadChars = 0;
unsigned int			g_RawWriteChars = 0;
unsigned int			g_PayloadReadChars = 0;
unsigned int			g_PayloadWriteChars = 0;
Threads::Mutex			g_logMtx;
#endif


/*************/
/* Functions */
/*************/


// Display usage.
void usage(char* exe)
{
	std::cerr << STRNAME << " version " << VERSION << " usage:" << std::endl;
	std::cerr << "   " << exe << " <config.file>" << std::endl;
}


// Parse the command line arguments and setup the global configuration
bool parseArgs(int argc, char** argv)
{
	bool ok = true;

	// Construct default config
	g_pConfig = new GlobalConfig();

	if(argc != 2) {
		ok = false;
	}
	else {
		try {
			// Create options parser to help with the parsing
			OptionsParser parser;

			// Get config from file
			char* cfgFile = argv[1];
			parser.parseCfgFile(cfgFile);
		}
		catch(ConfigError& e) {
			std::cerr << "Configuration error: " << e.what() << std::endl;
			ok = false;
		}

		#ifdef DEBUG_TUN
		// Setup debug settings
		if(g_pConfig) {
			g_pConfig->setDaemon(false);
		}
		#endif

		// Disable tunnel testing in this version
		if(g_pConfig) {
			g_pConfig->setTesting(false);
		}

		// Check we are setup as either a client or server
		if(ok && !g_pConfig->confCliSvr()) {
			std::cerr << "Configuration error: You must setup as either a ";
			std::cerr << "client or server" << std::endl;
			ok = false;
		}
	}

	return ok;
}


// Handler for SIGTERM/SIGINT
void signalHandler(void)
{
	// We will be waiting for SIGTERM/SIGINT
	sigset_t sigSet;

	#if defined(_POSIX_REALTIME_SIGNALS) && (_POSIX_REALTIME_SIGNALS != -1)
	sigemptyset(&sigSet);
	sigaddset(&sigSet, SIGTERM);
	sigaddset(&sigSet, SIGINT);
	sigaddset(&sigSet, SIGHUP);
	struct timespec sigTime = { 1, 0 };
	#endif

	// Just spin
	while(g_bRunning) {
		#if defined(_POSIX_REALTIME_SIGNALS) && (_POSIX_REALTIME_SIGNALS != -1)
		int sigNo = sigtimedwait(&sigSet, 0, &sigTime);
		if(sigNo == SIGTERM || sigNo == SIGINT) {
			g_bRunning = false;
		}
		else if(sigNo == SIGHUP) {
		}

		#else
		int sigNo = sigpending(&sigSet);
		if(sigNo == -1) {
			g_pLogger->error("sigpending() returned error");
			exit(1);
		}
		else if(sigismember(&sigSet, SIGTERM) || sigismember(&sigSet, SIGINT)) {
			g_bRunning = false;
		}
		else if(sigismember(&sigSet, SIGHUP)) {
		}
		else {
			sleep(1);
		}
		#endif
	}
}


// Turn us into a daemon
void daemonise(void)
{
	pid_t pid = fork();
	if(pid < 0) {
		g_pLogger->error("Fork() error");
		exit(1);
	}
	else if(pid != 0) {
		// Parent just exits
		exit(0);
	}

	setsid();
	umask(0);
}


#ifdef DEBUG_LOG_DATA
// Open debug log files
void openDebugLogs(void)
{
	// Set common filename stem
	std::string stem = g_pConfig->getDataLogFileStem();

	// Open log files
	std::string rawReadName = stem + "rawread.log";
	g_pRawReadLogFile = new std::ofstream(rawReadName.c_str());
	if(!g_pRawReadLogFile || !(*g_pRawReadLogFile)) {
		std::cerr << "Failed to open tunnel raw read log file" << std::endl;
		exit(1);
	}
	std::string rawWriteName = stem + "rawwrite.log";
	g_pRawWriteLogFile = new std::ofstream(rawWriteName.c_str());
	if(!g_pRawWriteLogFile || !(*g_pRawWriteLogFile)) {
		std::cerr << "Failed to open tunnel raw write log file" << std::endl;
		exit(1);
	}
	std::string payloadReadName = stem + "payloadread.log";
	g_pPayloadReadLogFile = new std::ofstream(payloadReadName.c_str());
	if(!g_pPayloadReadLogFile || !(*g_pPayloadReadLogFile)) {
		std::cerr << "Failed to open tunnel payload read log file" << std::endl;
		exit(1);
	}
	std::string payloadWriteName = stem + "payloadwrite.log";
	g_pPayloadWriteLogFile = new std::ofstream(payloadWriteName.c_str());
	if(!g_pPayloadWriteLogFile || !(*g_pPayloadWriteLogFile)) {
		std::cerr << "Failed to open tunnel payload write log file" << std::endl;
		exit(1);
	}
}
#endif


// Initialise general process functions
bool initProcess(int argc, char** argv)
{
	bool ok = true;

	// Construct logger
	#ifdef DEBUG_TUN
	g_pLogger = new Logger(STRNAME, true, false);
	#else
	g_pLogger = new Logger(STRNAME, false, true);
	#endif

	// Initialise SSL library
	try {
		SSLSock::initSSL();
	}
	catch(CommsError& e) {
		std::cerr << "Failed to initialise SSL library: " << e.what() << std::endl;
		exit(1);
	}

	// Parse args and setup config appropriately
	if(!parseArgs(argc, argv)) {
		usage(argv[0]);
		exit(1);
	}

	#ifdef DEBUG_LOG_DATA
	openDebugLogs();
	#endif

	// If requested, run as a daemon
	if(g_pConfig->getDaemon()) {
		daemonise();
	}

	// Create PID file
	std::string pidfilename = g_pConfig->getPIDFile();
	if(pidfilename.length() > 0) {
		std::ofstream pidfile(pidfilename.c_str());
		if(pidfile.good()) {
			pidfile << (unsigned int)getpid();
		}
		else {
			g_pLogger->warn("Failed to create PID file " + pidfilename);
		}
		pidfile.close();
	}

	// Block all signals
	sigset_t newmask;
	sigfillset(&newmask);
	if(pthread_sigmask(SIG_BLOCK, &newmask, 0) == -1) {
		g_pLogger->error("Error setting default signal mask");
		ok = false;
	}

	// Perform any final late configuration
	if(ok) {
		try {
			g_pConfig->doLateConfiguration();
		}
		catch(ConfigError& e) {
			std::cerr << "Late configuration error: " << e.what() << std::endl;
			ok = false;
		}
	}

	return ok;
}

// Main
int main(int argc, char** argv)
{
	// Initialise
	if(initProcess(argc, argv)) {
		// Do the tunneling
		try {
			// Kick off local servers
			g_pConfig->createLocalServers();

			// Run signal catcher in main thread
			signalHandler();
		}
		catch(CommsError& e) {
			g_pLogger->error("Fatal top level comms error: " + e.what());
		}
		catch(TunnelError& e) {
			g_pLogger->error("Fatal top level general error: " + e.what());
		}
		catch(Threads::ThreadError& e) {
			g_pLogger->error("Fatal top level thread error: " + e.what());
		}

		/* Tidy up and exit */

		// Remove PID file
		std::string pidfilename = g_pConfig->getPIDFile();
		if(pidfilename.length() > 0) {
			unlink(pidfilename.c_str());
		}

		// Free globals
		delete g_pConfig;
		g_pConfig = 0;
		delete g_pLogger;
		g_pLogger = 0;

		// Shutdown SSL library
		try {
			SSLSock::shutdownSSL();
		}
		catch(CommsError& e) {
			std::cerr << "Error closing SSL library: " << e.what() << std::endl;
		}
	}
	else {
		g_pLogger->error("Process initialisation failed");
	}

	exit(0);
}


